# -*- coding: utf-8 -*-

from collections import ChainMap
import os
from .metadataholder import MetaDataHolder
from .metadata import AbstractParameter, Parameter
from .datatypes import DataTypes
from .deserialize import Class_Look_Up

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

MdpInfo = {}


# These property names are not for mormal properties
Reserved_Property_Names = ['history', 'meta', 'refs', 'dataset',
                           'zInfo', '_MDP', 'extraMdp', 'alwaysMeta']


class Attributable(MetaDataHolder):
    """ An Attributable object is an object that has the
    notion of meta data.

    MetaData Porperties (MDPs) are Attributes that store their properties in te metadata table.
    """

    def __init__(self, meta=None, zInfo=None, alwaysMeta=False, **kwds):
        """ Pick out arguments listed in zInfo then put updated parameters into MetaData meta.

        meta: meta data container.
        zInfo: configuration
        alwaysMeta: always treat parameters as MetaDataProperties (MDPs).
        """

        self.alwaysMeta = alwaysMeta
        self.extraMdp = {}
        self.zInfo = zInfo if zInfo else {'metadata': {}}

        # super set of args and Model. Values only.
        # if an MDP is set to None in args, use it Model default
        mdps = {}
        zm = self.zInfo['metadata']
        for x in zm:
            # 'type' in args is 'typ_'. put `type` back.
            _x = 'typ_' if x == 'type' else x
            if _x not in kwds:
                mdps[x] = zm[x]['default']
            else:
                kwval = kwds.pop(_x)
                mdps[x] = zm[x]['default'] if kwval is None else kwval
        self._MDP = ChainMap(self.extraMdp, zm)
        super(Attributable, self).__init__(meta=meta, **kwds)
        self.setParameters(mdps)

    def setParameters(self, params):
        """ Set a group of name-value pairs to the object as properties.

        params: a dictionary of name:value where value is a subclass of
        `AbstractParameter`. value can be the value of a registered MDP.
        ``type`` will be used if ``typ_`` is given as the name.
        """

        for met, value in params.items():
            #  typ_ in params (from __init__) changed to type
            name = 'type' if met == 'typ_' else met
            # set to input if given or to default.
            self.__setattr__(name, value)
            #print('@@@@', name, value)

    @property
    def meta(self):
        return self.getMeta()

    @meta.setter
    def meta(self, newMetadata):
        self.setMeta(newMetadata)

    def setMeta(self, newMetadata):
        """ Replaces the current MetaData with specified argument.

        Product will override this to add listener when meta is
        replaced.
        `_defaults` which usually is `self.zInfo` is added to new meta so `str(meta)` can ommit the parameters with default value.
        """
        defs = self.zInfo['metadata'] if hasattr(self, 'zInfo') else None
        newMetadata._defaults = defs
        self._meta = newMetadata

    def __getattribute__(self, name):
        """ Returns the named metadata parameter.

        Reads meta data table when Attributes are
        read, and returns the values only.
        """
        # print('getattribute ' + name)

        # print('aa ' + selftr(self.getMeta()[name]))

        if name not in Reserved_Property_Names:
            try:
                if name in self._MDP:
                    return self._meta[name].getValue()
            except AttributeError:
                pass

        return super(Attributable, self).__getattribute__(name)

    def setMdp(self, name, value, met=None):

        m = self.getMeta()
        # print('MDP ', name, value, id(m), len(m))
        if name in m:
            # meta already has a Parameter for name
            p = m[name]
            if issubclass(value.__class__, AbstractParameter):
                tv, tp = value.getType(), p.getType()
                if issubclass(Class_Look_Up[DataTypes[tv]],
                              Class_Look_Up[DataTypes[tp]]):
                    p = value
                    return
                else:
                    vs = value.value
                    raise TypeError(
                        "Parameter %s type is %s, not %s's %s." % (vs, tv, name, tp))
            else:
                # value is not a Parameter
                v_type = type(value)
                p_type = Class_Look_Up[DataTypes[p.getType()]]
                if value is None or issubclass(v_type, p_type):
                    p.setValue(value)
                    return
                else:
                    vs = value
                    raise TypeError(
                        "Value %s type is %s, not %s's %s." % (vs, v_type.__name__, name, p_type.__name__))
        else:
            # named parameter is not in meta

            if issubclass(value.__class__, AbstractParameter):
                # value is a parameter
                m[name] = value
                return
            # value is not  a Parameter make one.
            m[name] = value2parameter(name, value, met)
        return

    def __setattr__(self, name, value):
        """ Stores value to attribute with name given.

        If name is in the `zInfo` list, store the value in a Parameter in
        metadata container. Updates meta data table. Updates value when
        an MDP attribute already has its Parameter in metadata.

        value: Must be Parameter/NumericParameter if this is normal metadata,
        depending on if it is `Number`. `Value` is the value if the  attribute
        is an MDP

        If ```self.alwaysMeta is True``` all properties of :class:`AbstractParameter` are taken as MDPs except those named in ```Reserved_Property_Names```
        """
        # print('setattr ' + name, value)
        if name in Reserved_Property_Names:
            super(Attributable, self).__setattr__(name, value)
            return
        try:
            if self.alwaysMeta:
                if issubclass(value.__class__, AbstractParameter):
                    # taken as an MDP attribute . store in meta
                    self.extraMdp[name] = value
                    self.setMdp(name, value, self._MDP)
                    # must return without updating self.__dict__
                    return
            if name in self._MDP:
                # an MDP attribute like 'description'. store in meta
                self.setMdp(name, value, self._MDP)
                # must return without updating self.__dict__
                return
        except AttributeError:
            pass

        super(Attributable, self).__setattr__(name, value)

    def __delattr__(self, name):
        """ Refuses deletion of mandatory attributes.
        """

        try:
            if self.alwaysMeta or \
               name in self.zInfo or \
               name in Reserved_Property_Names:
                logger.warn(
                    'Cannot delete MetaData Property or Reserved_Property_Names: ' + name)
                return
        except AttributeError:
            pass

        super(Attributable, self).__delattr__(name)

    def makeMdpMethods(self, outputfile=None):
        """ Generates a mix-in class file according to MetaData Property list.

        """
        filename = 'properties_' if outputfile is None else outputfile
        filename += self.__class__.__name__ + '.py'
        pth = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(pth, filename), 'w') as f:
            f.write('# -*- coding: utf-8 -*-\n')
            f.write(
                '\n### Automatically generated by %s.makeMdpMethods(). Do Not Edit.###' % self.__class__.__name__)
            f.write('\n\nclass MetaDataProperties():\n')
            s = make_class_properties(self.zInfo['metadata'].keys())
            f.write(s + '\n')


def value2parameter(name, value, descriptor):
    """ returns a parameter with correct type and attributes according to its value and name.

    value: type must be compatible with data_type. For example [0, 0] is wrong; Vector2d([0, 0)] is right if ``data_type``==``vector2d``.
    descriptor: is zInfo('metadata'] or zInfo['dataset'][xxx]
    """

    # if descriptor is None:
    #     im = {'description': 'UNKNOWN',
    #           'data_type': DataTypeNames[type(value).__name__],
    #           }
    # else:
    im = descriptor[name]  # {'dats_type':..., 'value':....}
    # in ['integer','hex','float','vector','quaternion']

    fs = im['default'] if 'default' in im else None
    gs = im['valid'] if 'valid' in im else None
    if im['data_type'] == 'string':
        from .stringparameter import StringParameter
        cs = im['typecode'] if 'typecode' in im else 'B'
        ret = StringParameter(value=value,
                              description=im['description'],
                              default=fs,
                              valid=gs,
                              typecode=cs
                              )
    elif im['data_type'] == 'finetime':
        from .dateparameter import DateParameter
        ret = DateParameter(value=value,
                            description=im['description'],
                            default=fs,
                            valid=gs,
                            )
    elif DataTypes[im['data_type']] in ['int', 'float', 'Vector', 'Vector2D', 'Quaternion']:
        from .numericparameter import NumericParameter
        us = im['unit'] if 'unit' in im else ''
        cs = im['typecode'] if 'typecode' in im else None
        ret = NumericParameter(value=value,
                               description=im['description'],
                               typ_=im['data_type'],
                               unit=us,
                               default=fs,
                               valid=gs,
                               typecode=cs,
                               )
    else:
        ret = Parameter(value=value,
                        description=im['description'],
                        typ_=im['data_type'],
                        default=fs,
                        valid=gs,
                        )
    return ret


def addMetaDataProperty(cls):
    """mh: Add MDP to a class so that although they are metadata,
    they can be accessed by for example, productfoo.creator.

    dynamic properties see
    https://stackoverflow.com/a/2584050
    https://stackoverflow.com/a/1355444
    """
    # MdpInfo is evaluated at class import time
    for name in MdpInfo:
        def g(self, n=name):
            return self._meta[name].getValue()

        def s(self, value, n=name):
            self.setMdp(n, value, MdpInfo)

        def d(self, n=name):
            logger.warn('Cannot delete MetaData Property ' + n)
        setattr(cls, name, property(
            g, s, d, 'MetaData Property ' + name))
    return cls


def make_class_properties(attrs, reserved=None):
    """ Generates class properties source code string according to given attribute names.

    """
    pr = []
    if reserved is None:
        reserved = Reserved_Property_Names
    for x in attrs:
        if x in reserved:
            raise NameError('"%s" is a reserved property name.' % x)
        arg = x        # x + '_' if x == 'type' else x
        pr.append('    @property')
        pr.append('    def %s(self): pass' % x)
        # pr.append('        return self._meta["%s"].getValue()\n' % x)
        # pr.append('    @%s.setter' % x)
        # pr.append('    def %s(self, p):' % x)
        # pr.append('        self.setMdp("%s", p, self._MDP)\n' % x)
    pr.append('    pass')
    properties = '\n'.join(pr)
    return properties


class MetaDataProperties():
    """ Mix-in place-holder for Class denifitions that has not their own `property.py` generated by :meth: `makeMdpMethods`.
    """

    pass



# -*- coding: utf-8 -*-
from os.path import join
import logging
import getpass
import os
from os.path import expanduser, expandvars

pnsconfig = {}

# look-up table for PoolManager (therefor HttpClient) to get pool URLs eith Pool ID (poolname)
poolurl_of = {
    'svom': 'http://10.0.10.114:9885/v0.6/svom'
}
pnsconfig['lookup'] = poolurl_of

###########################################
# Configuration for Servers running locally.
# components of the default poolurl

# the key (variable names) must be uppercased for Flask server
# FLASK_CONF = pnsconfig

# Te be edited automatically with sed -i 's/^EXTHOST =.*$/EXTHOST = xxx/g' file
EXTUSER = ''
EXTPASS = ''
EXTHOST = '172.17.0.1'
EXTPORT = 9876
MQUSER = ''
MQPASS = ''
MQHOST = '172.17.0.1'
MQPORT = 9876

BASE_POOLPATH = '/tmp'
SERVER_POOLPATH = '/tmp/data'

# base url for webserver. Update version if needed.
pnsconfig['api_version'] = 'v0.8'
pnsconfig['baseurl'] = '/' + pnsconfig['api_version']


# default path of working directories for LocalPools.
pnsconfig['base_poolpath'] = '/tmp'
# default path of working directories for HttpPool server.
pnsconfig['server_poolpath'] = '/tmp/data'

# For server
# you must have write permission of above paths.
# For example : /home/user/Documents
# will be added at the beginning of request pool ID 'my_pool'
# when you init a server pool like:
# ``pstore = PoolManager.getPool('my_pool') without poolurl,
# the poolurl will become 'file:///home/user/Documents/my_pool'.
# creating a pool at pc['base_poolpath']
# See document of :meth:'PoolManager.getPool'

# logging level for server
pnsconfig['logginglevel'] = logging.DEBUG
# for HttpPool
pnsconfig['defaultpool'] = 'default'

# message queue config
pnsconfig['mqtt'] = dict(
    host='x.x.x.x',
    port=31876,
    username='foo',
    passwd='bar',
)

# choose from pre-defined.
conf = ['dev', 'server_test', 'external'][0]

# modify
if conf == 'dev':
    # username, passwd, flask ip, flask port
    pnsconfig['node'] = {'username': 'foo',
                         'password': 'bar', 'host': '127.0.0.1', 'port': 5000}

    # server permission user
    pnsconfig['serveruser'] = 'mh'
    pnsconfig['base_poolpath'] = '/tmp'
    pnsconfig['server_poolpath'] = '/tmp/data'  # For server
    # PTS app permission user
    pnsconfig['ptsuser'] = 'mh'
    # on pns server
    home = '/home/' + pnsconfig['ptsuser']

elif conf == 'server_test':
    pnsconfig['node'] = {'username': 'foo', 'password': 'bar',
                         'host': '172.0.17.9', 'port': 9884}
    # server permission user
    pnsconfig['serveruser'] = 'apache'
    # PTS app permission user
    pnsconfig['ptsuser'] = 'pns'
    # on pns server
    home = '/home/' + pnsconfig['ptsuser']
elif conf == 'external':
    # wsgi behind apach2. cannot use env vars
    pnsconfig['node'] = {'username': 'foo', 'password': 'bar',
                         'host': EXTHOST, 'port': EXTPORT}
    # message queue config
    pnsconfig['mqtt'] = dict(
        host=MQHOST,
        port=MQPORT,
        username=MQUSER,
        passwd=MQPASS,
    )
    # server permission user
    pnsconfig['serveruser'] = 'apache'
    pnsconfig['base_poolpath'] = BASE_POOLPATH
    pnsconfig['server_poolpath'] = SERVER_POOLPATH  # For server
    # PTS app permission user
    pnsconfig['ptsuser'] = 'pns'
    # on pns server
    home = '/home/' + pnsconfig['ptsuser']
else:
    pass
pnsconfig['auth_user'] = pnsconfig['node']['username']
pnsconfig['auth_pass'] = pnsconfig['node']['password']
# It is best not use this 'httphost' in apps, and use url_of through getConfig
pnsconfig['httphost'] = 'http://' + \
    pnsconfig['node']['host']+':'+str(pnsconfig['node']['port'])
pnsconfig['mysql'] = {'host': 'ssa-mysql', 'port': 3306,
                      'user': 'root',  'password': '123456',
                      'database': 'users'}


# import user classes for server.
# See document in :class:`Classes`
pnsconfig['userclasses'] = ''


########### PNS-specific setup ############

phome = join(home, 'pns')
pnsconfig['paths'] = dict(
    pnshome=phome,
    inputdir=join(phome, 'input'),
    inputfiles=['pns.cat', 'pns.pn'],
    outputdir=join(phome, 'output'),
    outputfiles=['xycc.dat', 'atc.cc']
)

# the stateless data processing program that reads from inputdir and
# leave the output in the outputdir. The format is the input for subprocess()
h = pnsconfig['paths']['pnshome']
pnsconfig['scripts'] = dict(
    init=[join(h, 'initPTS'), ''],
    config=[join(h, 'configPTS'), ''],
    run=[join(h, 'runPTS'), ''],
    clean=[join(h, 'cleanPTS'), '']
)
del phome, h

# seconds
pnsconfig['timeout'] = 10



# -*- coding: utf-8 -*-

# Automatically generated from fdi/dataset/resources/TableDataset_DataModel.yml. Do not edit.

from collections import OrderedDict


from fdi.dataset.readonlydict import ReadOnlyDict

import copy

import logging
# create logger
logger = logging.getLogger(__name__)


# Data Model specification for mandatory components
_Model_Spec = {
    'name': 'TableDataset_DataModel',
    'description': 'TableDataset class data model mandatory configuration',
    'parents': [
        None,
        ],
    'schema': '1.6',
    'metadata': {
        'description': {
                'data_type': 'string',
                'description': 'Description of this dataset',
                'default': 'UNKNOWN',
                'valid': '',
                },
        'version': {
                'data_type': 'string',
                'description': 'Version of dataset',
                'default': '0.1',
                'valid': '',
                },
        'FORMATV': {
                'data_type': 'string',
                'description': 'Version of dataset schema and revision',
                'default': '1.6.0.1',
                'valid': '',
                },
        },
    'datasets': {
        },
    }

Model = ReadOnlyDict(_Model_Spec)





# -*- coding: utf-8 -*-

from .serializable import Serializable
from .deserialize import deserialize
from .eq import DeepEqual
from ..utils.common import trbk
from ..utils.getconfig import getConfig
from .listener import EventListener, EventSender
from .serializable import serialize
from ..utils.queueworks import queuework2

from collections import namedtuple, OrderedDict
from itertools import chain
import logging

if 0:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)8s %(process)d %(threadName)s %(levelname)s %(funcName)10s() %(lineno)3d- %(message)s')

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MqttRelayListener(EventListener):
    """ Generic interface for sending anything heard to an MQTT message queue.
    """

    def __init__(self, topics=None,
                 host=None, port=None, username=None, passwd=None,
                 callback=None, clean_session=None,
                 client_id=None, userdata=None,
                 qos=1, **kwds):
        """ Starts a MQTT message queue and forward everything in the arguement list to the MQ serialized.

        host, port, username, passwd: if any is not provided, it is looked up in `config['mqtt'].
        """
        super().__init__(**kwds)

        if bool(host and port and username and passwd) is False:
            conf = getConfig(conf='pns')

        self.mq = queuework2(
            topics,
            host=host if host else conf['mqtt']['host'],
            port=port if port else conf['mqtt']['port'],
            username=username if username else conf['mqtt']['username'],
            passwd=passwd if passwd else conf['mqtt']['passwd'],
            client_id=client_id, callback=callback,
            userdata=userdata if userdata else self,
            clean_session=clean_session, qos=qos)
        self.mq.logger = logger

    def targetChanged(self,  *args, **kwargs):
        """ Informs that an event has happened in a target of
        any type.
        """

        payload = list(chain(args, kwargs.items()))
        json_str = serialize(payload)

        logger.debug("send msg to [" + self.mq.topics + "]")
        logger.debug(json_str)
        if self.mq.start_send():
            if self.mq.send(self.mq.topics, json_str, conn=False) == 0:
                logger.debug('Send successfully')
                self.mq.stop_send()
        logger.debug("send over")


class MqttRelaySender(EventSender):
    """ Gets MQTT messages and forwards to listeners.

    """

    def __init__(self, topics=None,
                 host=None, port=None, username=None, passwd=None,
                 callback=None, clean_session=None,
                 client_id=None, userdata=None,
                 qos=1, **kwds):
        """ Starts a MQTT message queue and forward everything in the arguement list to the MQ serialized.

        host, port, username, passwd: if any is not provided, it is looked up in `config['mqtt'].
        """
        super().__init__(**kwds)

        if bool(host and port and username and passwd) is False:
            conf = getConfig(conf='pns')
        logger.debug('starting mq listening to '+str(topics))
        self.mq = queuework2(
            topics,
            host=host if host else conf['mqtt']['host'],
            port=port if port else conf['mqtt']['port'],
            username=username if username else conf['mqtt']['username'],
            passwd=passwd if passwd else conf['mqtt']['passwd'],
            client_id=client_id, callback=on_message,
            userdata=userdata if userdata else self,
            clean_session=clean_session, qos=qos)
        self.mq.logger = logger
        self.mq.start_receive(loop='start')


def on_message(client, userdata, msg):
    mqtt_rel_s = userdata
    logger.debug("Received: " + msg.topic + ' ' + str(msg.payload))

    msgobj = deserialize(msg.payload.decode(encoding='utf-8'))

    mqtt_rel_s.fire(msgobj)
    mqtt_rel_s.last_msg = msgobj
    client.loop_start()



# -*- coding: utf-8 -*-


from fdi.dataset.product import Product
from fdi.dataset.eq import deepcmp
from fdi.dataset.testproducts import get_sample_product
from fdi.pal.productstorage import ProductStorage
from fdi.pal.query import MetaQuery
from fdi.pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from fdi.pal.httpclientpool import HttpClientPool
from fdi.pns.fdi_requests import *
from fdi.utils.getconfig import getConfig
from fdi.utils.common import fullname

import pytest
import sys


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()


logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))


test_poolid = 'fdi_'+__name__


@pytest.fixture(scope="module")
def init_test():
    pass


def test_gen_url(setup):
    """ Makesure that request create corrent url
    """

    aburl, headers = setup
    samplepoolname = 'fdi_default_'+__name__
    samplepoolurl = aburl + '/' + samplepoolname
    sampleurn = 'urn:' + samplepoolname + ':fdi.dataset.product.Product:10'

    logger.info('Test GET HK')
    got_hk_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='housekeeping', method='GET')
    hk_url = aburl + '/' + samplepoolname + '/hk'
    assert got_hk_url == hk_url, 'Housekeeping url error: ' + got_hk_url + ':' + hk_url

    logger.info('Test GET classes, urns, tags, webapi url')
    got_classes_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='classes', method='GET')
    classes_url = aburl + '/' + samplepoolname + '/hk/classes'
    assert got_classes_url == classes_url, 'Classes url error: ' + got_classes_url

    got_urns_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='urns', method='GET')
    urns_url = aburl + '/' + samplepoolname + '/hk/urns'
    assert got_urns_url == urns_url, 'Urns url error: ' + got_urns_url

    got_tags_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='tags', method='GET')
    tags_url = aburl + '/' + samplepoolname + '/hk/tags'
    assert got_tags_url == tags_url, 'Housekeeping url error: ' + got_tags_url

    logger.info('Get product url')
    got_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='GET')
    product_url = aburl + '/'+samplepoolname + '/fdi.dataset.product.Product/10'
    assert got_product_url == product_url, 'Get product url error: ' + got_product_url

    logger.info('GET WebAPI  url')
    call = 'tagExists/foo'
    got_webapi_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents=call, method='GET')
    webapi_url = aburl + '/' + samplepoolname + '/' + 'api/' + call
    assert got_webapi_url == webapi_url, 'Get WebAPI url error: ' + got_webapi_url

    logger.info('Post product url')
    got_post_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='POST')
    post_product_url = aburl + '/'+samplepoolname + '/fdi.dataset.product.Product/10'
    assert got_post_product_url == post_product_url, 'Post product url error: ' + \
        got_post_product_url

    logger.info('Delete product url')
    got_del_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='DELETE')
    del_product_url = aburl + '/' + samplepoolname + '/fdi.dataset.product.Product/10'
    assert got_del_product_url == del_product_url, 'Delete product url error: ' + \
        got_del_product_url

    logger.info('Delete pool url')
    got_del_pool_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='pool', method='DELETE')
    del_pool_url = aburl + '/' + samplepoolname
    assert got_del_pool_url == del_pool_url, 'Delete product url error: ' + got_del_pool_url

    logger.info('Test corrupt request url')
    with pytest.raises(ValueError) as exc:
        err_url = urn2fdiurl(
            urn=sampleurn, poolurl=samplepoolurl, contents='pool', method='GET')
        exc_msg = exc.value.args[0]
        assert exc_msg == 'No such method and contents composition: GET/pool'


def test_CRUD_product_by_client(setup, local_pools_dir):
    """Client http product storage READ, CREATE, DELETE products in remote
    """
    aburl, headers = setup

    poolid = test_poolid
    poolurl = aburl + '/' + poolid
    pool = HttpClientPool(poolname=poolid, poolurl=poolurl)
    crud_t(poolid, poolurl, local_pools_dir, pool)


def crud_t(poolid, poolurl, local_pools_dir, pool):
    logger.info('Init a pstore')

    if PoolManager.isLoaded(DEFAULT_MEM_POOL):
        PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
    # this will also register the server side
    pstore = ProductStorage(pool=pool)
    pool.removeAll()

    assert len(pstore.getPools()) == 1, 'product storage size error: ' + \
        str(pstore.getPools())
    assert pstore.getPool(poolid) is not None, 'Pool ' + \
        poolid+' is None.'

    cnt = pool.getCount('fdi.dataset.product.Product')
    assert cnt == 0, 'Local metadata file size is 2'

    logger.info('Save data by ' + pool.__class__.__name__)
    x = Product(description='desc test')

    urn = pstore.save(x, geturnobjs=True)
    x.creator = 'httpclient'
    urn2 = pstore.save(x, geturnobjs=True)
    typenm = fullname(x)
    expected_urn = 'urn:' + poolid + ':' + fullname(x)
    assert urn.urn.rsplit(':', 1)[0] == expected_urn, \
        'Urn error: ' + expected_urn
    poolpath, scheme, place, pn = parse_poolurl(
        poolurl, poolhint=poolid)
    cnt = pool.getCount(typenm)
    assert cnt == 2

    logger.info('Load product from httpclientpool')
    res = pstore.getPool(poolid).loadProduct(urn2.urn)
    assert res.creator == 'httpclient', 'Load product error: ' + str(res)
    diff = deepcmp(x, res)
    assert diff is None, diff

    logger.info('Search metadata')
    q = MetaQuery(Product, 'm["creator"] == "httpclient"')
    res = pstore.select(q)
    assert len(res) == 1, 'Select from metadata error: ' + str(res)

    logger.info('Delete a product from httpclientpool')
    pstore.getPool(poolid).remove(urn.urn)
    lsn = pstore.getPool(poolid).getCount('fdi.dataset.product.Product')
    assert lsn == 1, 'Delete product local error, len sn : ' + lsn
    logger.info('A load exception message is expected')
    with pytest.raises(NameError):
        res = pstore.getPool(poolid).loadProduct(urn.urn)

    logger.info('Wipe a pool')
    pstore.getPool(poolid).removeAll()
    assert pool.isEmpty()

    tag = '==Sample ** Product=='
    logger.info('test sample prod with tag: '+tag)
    sp = get_sample_product()
    urn = pstore.save(sp, tag=tag)
    print('Sample Prod saved with tag "%s" %s to %s' %
          (tag, urn.urn, pool.poolname))

    logger.info('unregister a pool')
    assert len(pstore.getPools()) == 1, 'product storage size error: ' + \
        str(pstore.getPools())
    # unregister locally and remotely
    pstore.unregister(poolid)
    assert len(pstore.getPools()) == 0, 'product storage size error: ' + \
        str(pstore.getPools())

    logger.info('Access a non-existing pool and trgger an Error.')
    with pytest.raises(NameError):
        pstore.getPool(poolid) is None



# -*- coding: utf-8 -*-

from .productpool import ManagedPool
from ..utils.common import pathjoin, trbk

import filelock
import sys
import shutil
import mmap
import time
import json
import os
from os import path as op
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse, quote, unquote
else:
    PY3 = False
    strset = (str, unicode)
    from urlparse import urlparse, quote, unquote


def wipeLocal(path):
    """
    does the scheme-specific remove-all
    """
    # logger.debug()

    if path == '/':
        raise(ValueError('Do not remove root directory.'))

    if not op.exists(path):
        return
    try:
        shutil.rmtree(path)
        os.mkdir(path)
    except Exception as e:
        msg = 'remove-mkdir failed. exc: %s trbk: %s.' % (str(e), trbk(e))
        logger.error(msg)
        raise e


class LocalPool(ManagedPool):
    """ the pool will save all products in local computer.
    """

    def __init__(self, **kwds):
        """ creates file structure if there isn't one. if there is, read and populate house-keeping records. create persistent files if not exist.
        """
        # print(__name__ + str(kwds))
        super(LocalPool, self).__init__(**kwds)

    def setup(self):
        """ Sets up LocalPool interals.

        Make sure that self._poolname and self._poolurl are present.
        """

        if super().setup():
            return True

        real_poolpath = self.transformpath(self._poolname)
        if not op.exists(real_poolpath):
            os.makedirs(real_poolpath)
        self._files = {}
        self._atimes = {}
        self._cached_files = {}

        c, t, u = tuple(self.readHK().values())

        logger.debug('created ' + self.__class__.__name__ + ' ' + self._poolname +
                     ' at ' + real_poolpath + ' HK read.')

        self._classes.update(c)
        self._tags.update(t)
        self._urns.update(u)

        return False

    def readmmap(self, filename, close=False, check_time=False):
        fp = op.abspath(filename)
        if check_time:
            sr = os.stat(fp)
        if check_time and fp in self._atimes and (sr.st_mtime_ns <= self._atimes[fp]):
            # file hasnot changed since last time we read/wrote it.
            return None
        try:
            if 1:  # fp not in self._files or self._files[fp] is None:
                file_obj = open(fp, mode="r+", encoding="utf-8")
                # with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
            else:
                file_obj = self._files[fp]
                # file_obj.seek(0)
            js = file_obj.read()
        except Exception as e:
            msg = 'Error in HK reading. exc: %s trbk: %s.' % (str(e), trbk(e))
            logging.error(msg)
            raise NameError(msg)
        if 1:  # close:
            file_obj.close()
            if fp in self._files:
                del self._files[fp]
        else:
            self._files[fp] = file_obj
        if check_time:
            # save the mtime as the self atime
            self._atimes[fp] = sr.st_mtime_ns
        return js

    def readHK(self, hktype=None, serialize_out=False):
        """
        loads and returns the housekeeping data

        hktype: one of 'classes', 'tags', 'urns' to return. default is None to return alldirs
        serialize_out: if True return serialized form. Default is false.
        """
        if hktype is None:
            hks = ['classes', 'tags', 'urns']
        else:
            hks = [hktype]
        fp0 = self.transformpath(self._poolname)

        hk = {}
        for hkdata in hks:
            fp = op.abspath(pathjoin(fp0, hkdata + '.jsn'))
            if op.exists(fp):
                js = self.readmmap(fp, check_time=True)
                if js:
                    if serialize_out:
                        r = js
                    else:
                        from ..dataset.deserialize import deserialize
                        r = deserialize(js)
                    self._cached_files[fp] = js
                else:
                    # the file hasnot changed since last time we r/w it.
                    r = self._cached_files[fp] if serialize_out else \
                        self.__getattribute__('_' + hkdata)
            else:
                if serialize_out:
                    r = '{"_STID":"ODict"}'
                else:
                    from ..dataset.odict import ODict
                    r = ODict()
            hk[hkdata] = r
            assert r is not None
        logger.debug('HK read from ' + fp0)
        if serialize_out:
            return '{%s}' % ', '.join(('"%s": %s' % (k, v) for k, v in hk.items())) if hktype is None else hk[hktype]
        else:
            return hk if hktype is None else hk[hktype]

    def writeJsonmmap(self, fp, data, serialize_in=True, serialize_out=False, close=False, check_time=False, **kwds):
        """ write data in JSON from mmap file at fp.

        register the file. Leave file open by default `close`.
        data: to be serialized and saved.
        """
        from ..dataset.serializable import serialize

        js = serialize(data, **kwds) if serialize_in else data
        fp = op.abspath(fp)
        if 1:  # fp not in self._files or self._files[fp] is None:
            file_obj = open(fp, mode="w+", encoding="utf-8")
            # with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_WRITE) as mmap_obj:
        else:
            file_obj = self._files[fp]
        file_obj.seek(0)

        # file_obj.resize(len(js))
        file_obj.truncate(0)
        file_obj.write(js)
        # file_obj.flush()
        close = 1
        if close:
            file_obj.close()
            if fp in self._files:
                del self._files[fp]
        else:
            self._files[fp] = file_obj
        if check_time:
            # save the mtime as the self atime
            sr = os.stat(fp)
            os.utime(fp, ns=(sr.st_atime_ns, sr.st_mtime_ns))
            self._atimes[fp] = sr.st_mtime_ns
            self._cached_files[fp] = js
        l = len(js)
        logger.debug('JSON saved to: %s %d bytes' % (fp, l))
        return l

    def writeHK(self, fp0):
        """
           save the housekeeping data to disk
        """
        l = 0
        for hkdata in ['classes', 'tags', 'urns']:
            fp = pathjoin(fp0, hkdata + '.jsn')
            l += self.writeJsonmmap(fp, self.__getattribute__('_' + hkdata),
                                    check_time=True)
        return l

    def doSave(self, resourcetype, index, data, tag=None, serialize_in=True, **kwds):
        """
        does the media-specific saving.

        index: int
        """
        fp0 = self.transformpath(self._poolname)
        fp = pathjoin(fp0, quote(resourcetype) + '_' + str(index))
        try:
            # t0 = time.time()
            l = self.writeJsonmmap(
                fp, data, serialize_in=serialize_in, close=True, **kwds)
            l += self.writeHK(fp0)
            # print('tl %.8f %9d' % (time.time()-t0, l))
            logger.debug('HK written')
        except IOError as e:
            logger.error('Save failed. exc: %s trbk: %s.' % (str(e), trbk(e)))
            raise e  # needed for undoing HK changes
        return l

    def doLoad(self, resourcetype, index, serialize_out=False):
        """
        does the action of loading.

        """

        indexstr = str(index)
        pp = self.transformpath(self._poolname) + '/' + \
            resourcetype + '_' + indexstr
        js = self.readmmap(pp, close=True)
        if serialize_out:
            r = js
        else:
            from ..dataset.deserialize import deserialize
            r = deserialize(js)

        return r

    def doRemove(self, resourcetype, index):
        """
        does the action of removal of product from pool.
        """
        fp0 = self.transformpath(self._poolname)
        fp = op.abspath(pathjoin(fp0,  quote(resourcetype) + '_' + str(index)))
        try:
            if fp in self._files:
                if self._files[fp]:
                    self._files[fp].flush()
                    self._files[fp].close()
                del self._files[fp]
            os.unlink(fp)
            self.writeHK(fp0)
        except IOError as e:
            logger.error('Remove failed. exc: %s trbk: %s.' %
                         (str(e), trbk(e)))
            raise e  # needed for undoing HK changes

    def doWipe(self):
        """
        does the action of remove-all
        """
        for n, f in self._files.items():
            if f:
                f.flush()
                f.close()
        self._files.clear()
        self._atimes.clear()
        self._cached_files.clear()
        self._classes.clear()
        self._tags.clear()
        self._urns.clear()

        wipeLocal(self.transformpath(self._poolname))

    def getHead(self, ref):
        """ Returns the latest version of a given product, belonging
        to the first pool where the same track id is found.
        """

        raise(NotImplementedError())



# -*- coding: utf-8 -*-

# Automatically generated from fdi/dataset/resources/BaseProduct.yml. Do not edit.

from collections import OrderedDict
from fdi.dataset.finetime import FineTime


from fdi.dataset.readonlydict import ReadOnlyDict
from fdi.dataset.serializable import Serializable
from fdi.dataset.abstractcomposite import AbstractComposite
from fdi.dataset.listener import EventSender, EventType
from fdi.dataset.eq import deepcmp
from fdi.dataset.copyable import Copyable
from fdi.dataset.history import History

from collections import OrderedDict
import itertools

import logging
# create logger
logger = logging.getLogger(__name__)


class BaseProduct( AbstractComposite, Copyable, Serializable,  EventSender):
    """ A BaseProduct is the starting point of te whole  product tree, and a generic result that can be passed on between processes.

    In general a Product contains zero or more datasets, history,
    optional reference pointers and metadata some required metadata fields.
    A Product is expected to fully describe itself; this includes
    the way how this product was achieved (its history). As it is
    the result of a process, it should be able to save to and restore
    from an Archive device.

    Many times a Product may contain a single dataset and for this
    purpose the first dataset entry can be accessed by the getDefault()
    method. Note that the datasets may be a composite of datasets
    by themselves.

    A built-in attributes in `Model['metadata']` ("MetaData Parameter" or `MDP`) can be accessed with e.g. ``p.creator``, or p.meta['creator'].value::

        p.creator='foo'
    	assert p.creatur=='foo'
    	assert p.meta['creator']=='foo'
    	p.meta['creator']=Parameter('bar')
    	assert p.meta['creator']==Parameter('bar')

    =====
    BaseProduct class schema 1.6 inheriting [None].

Automatically generated from fdi/dataset/resources/BaseProduct.yml on 2021-06-18 16:19:33.951506.

Description:
FDI base class data model

    """

    def __init__(self,
                 description = 'UNKNOWN',
                 typ_ = 'BaseProduct',
                 level = 'ALL',
                 creator = 'UNKNOWN',
                 creationDate = FineTime(0),
                 rootCause = 'UNKNOWN',
                 version = '0.8',
                 FORMATV = '1.6.0.10',
                 zInfo=None,
                 **kwds):

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = OrderedDict(
            itertools.filterfalse(
                lambda x: x[0] in ('self', '__class__', 'zInfo', 'kwds'),
                locals().items())
        )

        global Model
	# instance variable for Model to be passed down inhritance chains.
        if zInfo is None:
            zInfo = Model

        # must be the first line to initiate meta and zInfo
        # :class: `Attributable` will process MDPs
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)

        self._history = History()


    @property
    def history(self):
        """ xx must be a property for ``self.xx = yy`` to work in super class after xx is set as a property also by a subclass.
        """
        return self._history

    @history.setter
    def history(self, history):
        self._history = history

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern."""
        visitor.visit(self)

    def getDefault(self):
        """ Convenience method that returns the first dataset \
        belonging to this product. """
        return list(self._sets.values())[0] if len(self._sets) > 0 else None



    def targetChanged(self, event):
        pass
        if event.source == self.meta:
            if event.type_ == EventType.PARAMETER_ADDED or \
               event.type_ == EventType.PARAMETER_CHANGED:
                # logger.debug(event.source.__class__.__name__ +   ' ' + str(event.change))
                pass

    def toString(self, level=0,
                 tablefmt='rst', tablefmt1='simple', tablefmt2='simple',
                 matprint=None, trans=True, beforedata='', **kwds):
        """ like AbstractComposite but with history
        """
        h = self.history.toString(
            level=level,
            tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
	    matprint=matprint, trans=trans, **kwds)
        s = super(BaseProduct, self).toString(
            level=level,
            tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
	    matprint=matprint, trans=trans, beforedata=h, **kwds)
        return s


    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        if 0:
            # remove self from meta's listeners because the deserialzed product will add itself during instanciation.
            print('1###' + self.meta.toString())
            metac = self.meta.copy()
            print('***' + metac.toString())
            print(deepcmp(self, metac.listeners[0]))
            metac.removeListener(self)

        ls = [
            ("meta", self.meta),
            ("_sets", self._sets if hasattr(self, '_sets') else None),
            ("history", self._history if hasattr(self, '_history') else None),
            ("listeners", self._listeners if hasattr(self, '_history') else None),
            ("_STID", self._STID if hasattr(self, '_STID') else None),
        ]

        return OrderedDict(ls)

    @property
    def description(self): pass
    @property
    def type(self): pass
    @property
    def level(self): pass
    @property
    def creator(self): pass
    @property
    def creationDate(self): pass
    @property
    def rootCause(self): pass
    @property
    def version(self): pass
    @property
    def FORMATV(self): pass
    pass

# Data Model specification for mandatory components
_Model_Spec = {
    'name': 'BaseProduct',
    'description': 'FDI base class data model',
    'parents': [
        None,
        ],
    'schema': '1.6',
    'metadata': {
        'description': {
                'id_zh_cn': '描述',
                'data_type': 'string',
                'description': 'Description of this product',
                'description_zh_cn': '对本产品的描述。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'type': {
                'id_zh_cn': '产品类型',
                'data_type': 'string',
                'description': 'Product Type identification. Name of class or CARD.',
                'description_zh_cn': '产品类型。完整Python类名或卡片名。',
                'default': 'BaseProduct',
                'valid': '',
                'typecode': 'B',
                },
        'level': {
                'id_zh_cn': '产品xx',
                'data_type': 'string',
                'description': 'Product level.',
                'description_zh_cn': '产品xx',
                'default': 'ALL',
                'valid': '',
                'typecode': 'B',
                },
        'creator': {
                'id_zh_cn': '本产品生成者',
                'data_type': 'string',
                'description': 'Generator of this product.',
                'description_zh_cn': '本产品生成方的标识，例如可以是单位、组织、姓名、软件、或特别算法等。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'creationDate': {
                'id_zh_cn': '产品生成时间',
                'fits_keyword': 'DATE',
                'data_type': 'finetime',
                'description': 'Creation date of this product',
                'description_zh_cn': '本产品生成时间',
                'default': 0,
                'valid': '',
                'typecode': None,
                },
        'rootCause': {
                'id_zh_cn': '数据来源',
                'data_type': 'string',
                'description': 'Reason of this run of pipeline.',
                'description_zh_cn': '数据来源（此例来自鉴定件热真空罐）',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'version': {
                'id_zh_cn': '版本',
                'data_type': 'string',
                'description': 'Version of product',
                'description_zh_cn': '产品版本',
                'default': '0.8',
                'valid': '',
                'typecode': 'B',
                },
        'FORMATV': {
                'id_zh_cn': '格式版本',
                'data_type': 'string',
                'description': 'Version of product schema and revision',
                'description_zh_cn': '产品格式版本',
                'default': '1.6.0.10',
                'valid': '',
                'typecode': 'B',
                },
        },
    'datasets': {
        },
    }

Model = ReadOnlyDict(_Model_Spec)

MdpInfo = Model['metadata']




#!/usr/bin/env python
"""Get TAI-UTC difference in seconds for a given time using tzdata.

i.e., find the number of seconds that must be added to UTC to compute
TAI for any timestamp at or after the given time[1].

  >>> from datetime import datetime
  >>> import leapseconds
  >>> leapseconds.dTAI_UTC_from_utc(datetime(2005, 1, 1, tzinfo = timezone.utc))
  datetime.timedelta(0, 32)
  >>> leapseconds.utc_to_tai(datetime(2015, 7, 1, tzinfo = timezone.utc))
  datetime.datetime(2015, 7, 1, 0, 0, 36, tzinfo=datetime.timezone.utc)
  >>> leapseconds.tai_to_utc(datetime(2015, 7, 1, 0, 0, 36, tzinfo = timezone.utc))
  datetime.datetime(2015, 7, 1, 0, 0, tzinfo=datetime.timezone.utc)
  >>> leapseconds.tai_to_utc(datetime(2015, 7, 1, 0, 0, 35, tzinfo = timezone.utc)) # leap second
  datetime.datetime(2015, 7, 1, 0, 0, tzinfo=datetime.timezone.utc)
  >>> leapseconds.tai_to_utc(datetime(2015, 7, 1, 0, 0, 34, tzinfo = timezone.utc))
  datetime.datetime(2015, 6, 30, 23, 59, 59, tzinfo=datetime.timezone.utc)

Python 2.6+, Python 3, Jython, Pypy support.

[1]: https://github.com/eggert/tz/blob/master/leap-seconds.list
[2]: https://github.com/eggert/tz/blob/master/tzfile.h
[3]: https://github.com/eggert/tz/blob/master/zic.c
[4]: http://datacenter.iers.org/eop/-/somos/5Rgv/latest/16
"""
from __future__ import with_statement

from collections import namedtuple
from datetime import datetime, timedelta, timezone
from struct import Struct
from warnings import warn
from functools import partial, lru_cache

__all__ = ['leapseconds', 'LeapSecond',
           'dTAI_UTC_from_utc', 'dTAI_UTC_from_tai',
           'tai_to_utc', 'utc_to_tai',
           'gps_to_utc', 'utc_to_gps',
           'tai_to_gps', 'gps_to_tai']

__version__ = "0.1.0"

# from timezone/tzfile.h [2] (the file is in public domain)
"""
struct tzhead {
 char tzh_magic[4];  /* TZ_MAGIC */
 char tzh_version[1];  /* '\0' or '2' or '3' as of 2013 */
 char tzh_reserved[15]; /* reserved--must be zero */
 char tzh_ttisgmtcnt[4]; /* coded number of trans. time flags */
 char tzh_ttisstdcnt[4]; /* coded number of trans. time flags */
 char tzh_leapcnt[4];  /* coded number of leap seconds */
 char tzh_timecnt[4];  /* coded number of transition times */
 char tzh_typecnt[4];  /* coded number of local time types */
 char tzh_charcnt[4];  /* coded number of abbr. chars */
};

# from zic.c[3] (the file is in public domain)
convert(const int_fast32_t val, char *const buf)
{
        register int i;
        register int shift;
        unsigned char *const b = (unsigned char *) buf;

        for (i = 0, shift = 24; i < 4; ++i, shift -= 8)
                b[i] = val >> shift;
}
# val = 0x12345678
# (val >> 24) & 0xff, (val >> 16) & 0xff, (val >> 8) & 0xff, val & 0xff
# 0x12 0x34 0x56 0x78
# therefore "coded number" means big-endian 32-bit integer
"""

dTAI_GPS = timedelta(seconds=19)  # constant offset

LeapSecond = namedtuple('LeapSecond', 'utc dTAI_UTC')  # tai = utc + dTAI_UTC
sentinel = LeapSecond(utc=datetime.max.replace(
    tzinfo=timezone.utc), dTAI_UTC=timedelta(0))


def leapseconds(tzfiles=['/usr/share/zoneinfo/right/UTC',
                         '/usr/lib/zoneinfo/right/UTC'],
                use_fallback=True):
    """Extract leap seconds from *tzfiles*."""
    for filename in tzfiles:
        try:
            file = open(filename, 'rb')
        except IOError:
            continue
        else:
            break
    else:  # no break
        if not use_fallback:
            raise ValueError('Unable to open any tzfile: %s' % (tzfiles,))
        else:
            return _fallback()

    with file:
        header = Struct('>4s c 15x 6i')  # see struct tzhead above
        (magic, version, _, _, leapcnt, timecnt, typecnt,
         charcnt) = header.unpack_from(file.read(header.size))
        if magic != "TZif".encode():
            if not use_fallback:
                raise ValueError('Wrong magic %r in tzfile: %s' % (
                    magic, file.name))
            else:
                return _fallback()
        if version not in '\x0023'.encode():
            warn('Unsupported version %r in tzfile: %s' % (
                version, file.name), RuntimeWarning)
        if leapcnt == 0:
            if not use_fallback:
                raise ValueError("No leap seconds in tzfile: %s" % (
                    file.name))
            else:
                return _fallback()

        """# from tzfile.h[2] (the file is in public domain)

         . . .header followed by. . .

         tzh_timecnt (char [4])s  coded transition times a la time(2)
         tzh_timecnt (unsigned char)s types of local time starting at above
         tzh_typecnt repetitions of
           one (char [4])  coded UT offset in seconds
           one (unsigned char) used to set tm_isdst
           one (unsigned char) that's an abbreviation list index
         tzh_charcnt (char)s  '\0'-terminated zone abbreviations
         tzh_leapcnt repetitions of
           one (char [4])  coded leap second transition times
           one (char [4])  total correction after above
        """
        file.read(timecnt * 5 + typecnt * 6 + charcnt)  # skip

        result = [LeapSecond(
            datetime(1972, 1, 1, tzinfo=timezone.utc), timedelta(seconds=10))]
        nleap_seconds = 10
        tai_epoch_as_tai = datetime(1970, 1, 1, 0, 0, 10, tzinfo=timezone.utc)
        buf = Struct(">2i")
        for _ in range(leapcnt):  # read leap seconds
            t, cnt = buf.unpack_from(file.read(buf.size))
            dTAI_UTC = nleap_seconds + cnt
            utc = tai_epoch_as_tai + timedelta(seconds=t - dTAI_UTC + 1)
            assert utc - datetime(utc.year, utc.month,
                                  utc.day, tzinfo=timezone.utc) == timedelta(0)
            result.append(LeapSecond(utc, timedelta(seconds=dTAI_UTC)))
        result.append(sentinel)
        return result


@lru_cache(maxsize=8)
def _fallback():
    """Leap seconds list if no tzfiles are available."""
    td = timedelta
    dt_utc = partial(datetime, tzinfo=timezone.utc)
    return [
        # appropriate
        LeapSecond(utc=dt_utc(1958, 1, 1, 0, 0), dTAI_UTC=td(0, 0)),
        LeapSecond(utc=dt_utc(1961, 1, 1, 0, 0), dTAI_UTC=td(0, 1.442818)),
        LeapSecond(utc=dt_utc(1961, 8, 1, 0, 0), dTAI_UTC=td(0, 1.372818)),
        LeapSecond(utc=dt_utc(1962, 1, 1, 0, 0), dTAI_UTC=td(0, 1.845858)),
        LeapSecond(utc=dt_utc(1963, 11, 1, 0, 0), dTAI_UTC=td(0, 1.945858)),
        LeapSecond(utc=dt_utc(1964, 1, 1, 0, 0), dTAI_UTC=td(0, 3.240130)),
        LeapSecond(utc=dt_utc(1964, 4, 1, 0, 0), dTAI_UTC=td(0, 3.340130)),
        LeapSecond(utc=dt_utc(1964, 9, 1, 0, 0), dTAI_UTC=td(0, 3.440130)),
        LeapSecond(utc=dt_utc(1965, 1, 1, 0, 0), dTAI_UTC=td(0, 3.540130)),
        LeapSecond(utc=dt_utc(1965, 3, 1, 0, 0), dTAI_UTC=td(0, 3.640130)),
        LeapSecond(utc=dt_utc(1965, 7, 1, 0, 0), dTAI_UTC=td(0, 3.740130)),
        LeapSecond(utc=dt_utc(1965, 9, 1, 0, 0), dTAI_UTC=td(0, 3.840130)),
        LeapSecond(utc=dt_utc(1966, 1, 1, 0, 0), dTAI_UTC=td(0, 4.313170)),
        LeapSecond(utc=dt_utc(1968, 2, 1, 0, 0), dTAI_UTC=td(0, 4.213170)),

        LeapSecond(utc=dt_utc(1972, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 10)),
        LeapSecond(utc=dt_utc(1972, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 11)),
        LeapSecond(utc=dt_utc(1973, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 12)),
        LeapSecond(utc=dt_utc(1974, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 13)),
        LeapSecond(utc=dt_utc(1975, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 14)),
        LeapSecond(utc=dt_utc(1976, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 15)),
        LeapSecond(utc=dt_utc(1977, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 16)),
        LeapSecond(utc=dt_utc(1978, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 17)),
        LeapSecond(utc=dt_utc(1979, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 18)),
        LeapSecond(utc=dt_utc(1980, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 19)),
        LeapSecond(utc=dt_utc(1981, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 20)),
        LeapSecond(utc=dt_utc(1982, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 21)),
        LeapSecond(utc=dt_utc(1983, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 22)),
        LeapSecond(utc=dt_utc(1985, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 23)),
        LeapSecond(utc=dt_utc(1988, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 24)),
        LeapSecond(utc=dt_utc(1990, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 25)),
        LeapSecond(utc=dt_utc(1991, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 26)),
        LeapSecond(utc=dt_utc(1992, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 27)),
        LeapSecond(utc=dt_utc(1993, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 28)),
        LeapSecond(utc=dt_utc(1994, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 29)),
        LeapSecond(utc=dt_utc(1996, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 30)),
        LeapSecond(utc=dt_utc(1997, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 31)),
        LeapSecond(utc=dt_utc(1999, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 32)),
        LeapSecond(utc=dt_utc(2006, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 33)),
        LeapSecond(utc=dt_utc(2009, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 34)),
        LeapSecond(utc=dt_utc(2012, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 35)),
        LeapSecond(utc=dt_utc(2015, 7, 1, 0, 0), dTAI_UTC=timedelta(0, 36)),
        LeapSecond(utc=dt_utc(2017, 1, 1, 0, 0), dTAI_UTC=timedelta(0, 37)),
        sentinel]


def dTAI_UTC_from_utc(utc_time):
    """TAI time = utc_time + dTAI_UTC_from_utc(utc_time)."""
    return _dTAI_UTC(utc_time, lambda ls: ls.utc)


def dTAI_UTC_from_tai(tai_time):
    """UTC time = tai_time - dTAI_UTC_from_tai(tai_time)."""
    return _dTAI_UTC(tai_time, lambda ls: ls.utc + ls.dTAI_UTC)


def _dTAI_UTC(time, leapsecond_to_time, leapseconds=leapseconds):
    """Get TAI-UTC difference in seconds for a given time.

    >>> from datetime import datetime, timedelta
    >>> _dTAI_UTC(datetime(1972, 1, 1, tzinfo = timezone.utc), lambda ls: ls.utc)
    datetime.timedelta(0, 10)
    >>> tai = lambda ls: ls.utc + ls.dTAI_UTC
    >>> _dTAI_UTC(datetime(2015, 7, 1, 0, 0, 34, tzinfo = timezone.utc), tai)
    datetime.timedelta(0, 35)
    >>> _dTAI_UTC(datetime(2015, 7, 1, 0, 0, 35, tzinfo = timezone.utc), tai) # leap second
    datetime.timedelta(0, 35)
    >>> _dTAI_UTC(datetime(2015, 7, 1, 0, 0, 36, tzinfo = timezone.utc), tai)
    datetime.timedelta(0, 36)

    Bulletin C 51 says "NO leap second will be introduced at the end
    of June 2016."[4] and therefore UTC-TAI is still 36
    at 27 June 2016:

    >>> _dTAI_UTC(datetime(2016, 6, 27, tzinfo = timezone.utc), lambda ls: ls.utc)
    datetime.timedelta(0, 36)
    """
    leapseconds_list = leapseconds()
    transition_times = list(map(leapsecond_to_time, leapseconds_list))
    if time < transition_times[0]:
        # try fallback first
        leapseconds_list = _fallback()
        transition_times = list(map(leapsecond_to_time, leapseconds_list))
        if time < transition_times[0]:
            raise ValueError("Dates before %s are not supported, got %r" % (
                transition_times[0], time))
    for i, (start, end) in enumerate(zip(transition_times,
                                         transition_times[1:])):
        if start <= time < end:
            return leapseconds_list[i].dTAI_UTC
    assert 0


def tai_to_utc(tai_time):
    """Convert TAI time given as datetime object to UTC time."""
    return tai_time - dTAI_UTC_from_tai(tai_time)


def utc_to_tai(utc_time):
    """Convert UTC time given as datetime object to TAI time."""
    return utc_time + dTAI_UTC_from_utc(utc_time)


def gps_to_utc(gps_time):
    """Convert GPS time given as datetime object to UTC time."""
    return tai_to_utc(gps_to_tai(gps_time))


def utc_to_gps(utc_time):
    """Convert UTC time given as datetime object to GPS time."""
    return tai_to_gps(utc_to_tai(utc_time))


def tai_to_gps(tai_time):
    """Convert TAI time given as datetime object to GPS time."""
    return tai_time - dTAI_GPS


def gps_to_tai(gps_time):
    """Convert GPS time given as datetime object to TAI time."""
    return gps_time + dTAI_GPS


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    import json

    assert all(ls.dTAI_UTC == timedelta(seconds=ls.dTAI_UTC.seconds,
                                        microseconds=ls.dTAI_UTC.microseconds)
               for ls in leapseconds())  # ~+200 leap second until 2100
    print(json.dumps([dict(utc=t.utc, tai=t.utc + t.dTAI_UTC,
                           dTAI_UTC=t.dTAI_UTC.seconds)
                      for t in leapseconds()],
                     default=str, indent=4, sort_keys=True))



# -*- coding: utf-8 -*-

from ..utils.common import mstr, bstr
from .listener import MetaDataListener
from .serializable import Serializable
from .typed import Typed
from .odict import ODict
from .attributable import Attributable
from .abstractcomposite import AbstractComposite
from .datawrapper import DataContainer, DataWrapper
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable


from collections import OrderedDict
import logging
import sys

if sys.version_info[0] + 0.1 * sys.version_info[1] >= 3.3:
    PY33 = True
    from collections.abc import Container, Sequence, Mapping
    seqlist = Sequence
    maplist = Mapping
else:
    assert 0, 'python 3'
    PY33 = False
    from .collectionsMockUp import ContainerMockUp as Container
    from .collectionsMockUp import SequenceMockUp as Sequence
    from .collectionsMockUp import MappingMockUp as Mapping
    seqlist = (tuple, list, Sequence, str)
    # ,types.XRangeType, types.BufferType)
    maplist = (dict, Mapping)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

# from .composite import


class Dataset(Attributable, DataContainer, Serializable, MetaDataListener):
    """ Attributable and annotatable information data container
    that can be be part of a Product.

    developer notes:
    The intent is that developers do not derive from this interface

    directly. Instead, they should inherit from one of the generic
    datasets that this package provides:

    mh: GenericDataset,
    ArrayDataset.
    TableDataset or
    CompositeDataset.
    """

    def __init__(self, **kwds):
        """

        """
        super(Dataset, self).__init__(**kwds)

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern."""
        visitor.visit(self)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """

        s = OrderedDict(description=self.description,
                        meta=self._meta,
                        data=self.data,
                        _STID=self._STID)  # TODO
        return s


class GenericDataset(Dataset, Typed, DataWrapper):
    """ mh: Contains one typed data item with a unit and a typecode.
    """

    def __init__(self,                 **kwds):
        """
        """
        super(GenericDataset, self).__init__(
            **kwds)  # initialize data, meta, unit

    def __iter__(self):
        for x in self.getData():
            yield x

    def toString(self, level=0,
                 tablefmt='rst', tablefmt1='simple', tablefmt2='simple',
                 param_widths=None, width=0, matprint=None, trans=True, **kwds):
        """ matprint: an external matrix print function
        trans: print 2D matrix transposed. default is True.
        """
        cn = self.__class__.__name__
        if level > 1:
            return cn + \
                '{ %s, description = "%s", meta = %s }' % \
                (str(self.data), str(self.description), self.meta.toString(
                    tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                    level=level, width=width, param_widths=param_widths,
                    matprint=matprint, trans=trans, **kwds))

        s = '=== %s (%s) ===\n' % (cn, self.description if hasattr(
            self, 'descripion') else '')
        s += mstr(self.__getstate__(), level=level,
                  tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                  excpt=['description'], **kwds)

        d = cn + '-dataset =\n'
        d += bstr(self.data, level=level,
                  tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                  **kwds) if matprint is None else \
            matprint(self.data, level=level, trans=False, headers=[], tablefmt2='plain',
                     **kwds)
        return s + '\n' + d + '\n'


class CompositeDataset(AbstractComposite, Dataset):
    """  An CompositeDataset is a Dataset that contains zero or more
    named Datasets. It allows to build arbitrary complex dataset
    structures.

    It also defines the iteration ordering of its children, which is
    the order in which the children were inserted into this dataset.
    """

    def __init__(self, **kwds):
        """
        """
        super(CompositeDataset, self).__init__(
            **kwds)  # initialize _sets, meta, unit

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           meta=self.meta,
                           _sets=self._sets,
                           _STID=self._STID)



# -*- coding: utf-8 -*-

from .indexed import Indexed
from .ndprint import ndprint
from .odict import ODict
from ..utils.common import mstr, bstr, lls, exprstrs
from .dataset import Dataset
from .indexed import Indexed

try:
    from .tabledataset_datamodel import Model
except ImportError:
    Model = {'metadata': {}}


import sys
from collections.abc import Sequence
from collections import OrderedDict
import itertools

if sys.version_info[0] + 0.1 * sys.version_info[1] >= 3.3:
    PY33 = True
    from collections.abc import Container, Sequence, Mapping
    seqlist = Sequence
    maplist = Mapping
else:
    assert 0, 'python 3'
    PY33 = False
    from .collectionsMockUp import ContainerMockUp as Container
    from .collectionsMockUp import SequenceMockUp as Sequence
    from .collectionsMockUp import MappingMockUp as Mapping
    seqlist = (tuple, list, Sequence, str)
    # ,types.XRangeType, types.BufferType)
    maplist = (dict, Mapping)

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class TableModel(object):
    """ to interrogate a tabular data model
    """

    def __init__(self, **kwds):
        """

        """
        super(TableModel, self).__init__(**kwds)

    def getColumnClass(self, columnIndex):
        """ Returns the class for the first cell
        values in the column.
        """
        return self.getColumn(columnIndex)[0].__class__

    def getColumnCount(self):
        """ Returns the number of columns in the model. """
        return len(self.getData())

    def getColumnName(self, columnIndex):
        """ Returns the name of the column at columnIndex.

        returns a set of columns if key  is a slice.
        """

        return self.getColumnNames()[columnIndex]

    def getColumnNames(self):
        """ Returns the column names. """
        return list(self.getData().keys())

    def getRowCount(self):
        """ Returns the number of rows in the model. """
        return len(self.getColumn(0))

    def getValueAt(self, rowIndex, columnIndex):
        """ Returns the value for the cell at columnIndex and rowIndex. """
        return self.getColumn(columnIndex).data[rowIndex]

    def isCellEditable(self, rowIndex, columnIndex):
        """ Returns true if the cell at rowIndex and columnIndex
        is editable. """
        return True

    def setValueAt(self, value, rowIndex, columnIndex):
        """Sets the value in the cell at columnIndex and rowIndex
        to Value.
        """
        self.getColumn(columnIndex).data[rowIndex] = value


MdpInfo = Model['metadata']


class TableDataset(Dataset, TableModel):
    """  Special dataset that contains a single Array Data object.
    A TableDataset is a tabular collection of Columns. It is optimized to work on array data..
    The column-wise approach is convenient in many cases. For example, one has an event list, and each algorithm is adding a new field to the events (i.e. a new column, for example a quality mask).

    Although mechanisms are provided to grow the table row-wise, one should use these with care especially in performance driven environments as this orthogonal approach (adding rows rather than adding columns) is expensive.

    General Note:

    For reasons of flexibility, memory consumption and performance, this class is not checking whether all columns are of the same length: this is the responsibility of the user/developer. See also the library documentation for more information about this.

    Note on column names:

    If a column is added without specifying a name, the name ColumnX is created, where X denotes the index of that column.
    Column name duplicity is not allowed.

    Developers:

    See "Writing special datasets or products" at the developer's documentation also.


    Please see also this selection example.
    """

    def __init__(self, data=None,
                 zInfo=None,
                 alwaysMeta=True,
                 **kwds):
        """
        """

        self._list = []

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = OrderedDict(
            itertools.filterfalse(
                lambda x: x[0] in ('self', '__class__', 'zInfo', 'kwds'),
                locals().items())
        )

        global Model
        if zInfo is None:
            zInfo = Model

        super().__init__(zInfo=zInfo, **metasToBeInstalled,
                         **kwds)  # initialize data, meta, unit
        # self.setData(data)

    def getData(self):
        """ Optimized for _data being an ``ODict`` implemented with ``UserDict``.

        """

        try:
            return self._data.data
        except AttributeError:
            d = super().getData()
            if hasattr(d, 'data'):
                return self._data.data
            else:
                return d

    def setData(self, data):
        """ sets name-column pairs from data.

        Valid formd include: {str:Column, ...} or [(str, [num, ...], str)]
        or [(str, Column), ...] or [[num ...],  [num ...], ...]

        [{'name':str,'column':Column}] form is deprecated.

        Existing data will be discarded except when the provided data is a list of lists, where existing column names and units will remain but data replaced, and extra data items will form new columns named 'column'+index (index counting from 1) with unit None.
        """
        # logging.debug(data.__class__)

        if data is None:
            super(TableDataset, self).setData(ODict())
            return

        current_data = self.getData()
        # list of keys of current data
        curdk = list(current_data.keys()) if current_data else []
        super(TableDataset, self).setData(ODict())
        if issubclass(data.__class__, seqlist):
            from .arraydataset import Column
            for ind, x in enumerate(data):
                if issubclass(x.__class__, maplist) \
                   and 'name' in x and 'column' in x:
                    raise DeprecationWarning(
                        'Do not use [{"name":name, "column":column}...]. Use {name:column, ...} instead.')
                if issubclass(x.__class__, (list, tuple)):
                    # check out string-started columns
                    if len(x) > 1 and issubclass(x[0].__class__, str) and not issubclass(x[1].__class__, str):
                        if issubclass(x[1].__class__, (list, tuple)):
                            u = x[2] if len(x) > 2 else ''
                            self.setColumn(x[0], Column(data=x[1], unit=x[2]))
                        elif issubclass(x[1].__class__, Column):
                            self.setColumn(x[0], x[1])
                        else:
                            raise '[[str, [], str]...], [[str, []]...], [[str, Column]...] needed.'
                    else:
                        if current_data is None or len(current_data) <= ind:
                            # update the data of the ind-th column
                            self.setColumn('', Column(data=x, unit=None))
                        else:
                            colname = curdk[ind]
                            current_data[colname].data = x
                            self.setColumn(colname, current_data[colname])
                else:
                    raise ValueError(
                        'Cannot extract name and column from list member ' + str(x))
        elif issubclass(data.__class__, maplist):
            for k, v in data.items():
                self.setColumn(k, v)
        else:
            raise TypeError('must be a Sequence or a Mapping. ' +
                            data.__class__.__name__ + ' found.')

    def addColumn(self, name, column, col_des=False):
        """ Adds the specified column to this table, and attaches a name
        to it. If the name is null, a dummy name "column"+column_count+1 is created, such that it can be accessed by getColumn(str).

        If column name exists the corresponding column is substituted.

        Parameters:
        name - column name.
        column - column to be added.
        col_des - if True (default) and column descripion is 'UNKNOWN', set to column name.
        """

        d = self.getData()
        if d is None:
            d = ODict()

        if name == '' or name is None:
            idx = self.getColumnCount()
            name = 'column' + str(idx+1)
            self._list.append(column.getData())
        else:
            try:
                self._list[self.indexOf(name)] = column.getData()
            except ValueError as e:
                self._list.append(column.getData())
        if col_des and column.getDescription() == 'UNKNOWN':
            column.setDescription(name)
        d[name] = column

    def removeColumn(self, key):
        """ Removes the columns specified by ``key``.

        ref. ``getColumnMap`` on ``key`` usage.
        """

        for name in self.getColumnMap(key).keys():
            self._list.pop(self.indexOf(name))
            del(self.data[name])

    def indexOf(self, key):
        """ Returns the index of specified column.

        if the key is a Column,
        it looks for equal references (same column objects), not for
        equal values.
        If the key is a string, Returns the index of specified Column name.
        """
        from .arraydataset import Column
        if issubclass(key.__class__, str):
            ks = list(self.getData().keys())
            k = key
        elif issubclass(key.__class__, Column):
            ks = list(id(v) for v in self.getData().values())
            k = id(key)
        else:
            raise "key must be string or Column, not %s." % type(key).__name__

        return ks.index(k)

    def addRow(self, row, rows=False):
        """ Adds the specified map as a new row to this table.

        row: mh: row is a dict with names as keys and row data as value.
        rows: append each element in row if the row data is a list.
        """

        d = self.getData()
        if len(row) < len(d):
            msg = 'row width d% should be %d.' % (len(row), len(d))
            raise ValueError(msg)

        for c in d.keys():
            if rows:
                d[c].data.extend(row[c])
            else:
                d[c].data.append(row[c])

    def getRowMap(self, rowIndex):
        """ Returns a dict of column-names as the keys and the objects located at a particular row(s) as the values.

        rowIndex: return the following as the value for each key-value pair:
* int: the int-th row's elements;
* ``Slice`` object, a list of rows from slicing the column. Example ``a.getRow(Slice(3,,))``;
* list of integers: they are used as the row index to select the rows.
* list of booleans: rows where the corresponding boolean is True are chosen.
        """
        cl = rowIndex.__class__
        d = self.getData()
        if issubclass(cl, (int, slice)):
            return {n: c.getData()[rowIndex] for n, c in d.items()}
        if issubclass(cl, list):
            if type(rowIndex[0]) == int:
                return {n: [c.getData()[i] for i in rowIndex] for n, c in d.items()}
            if type(rowIndex[0]) == bool:
                # if len(rowIndex) != len(n):
                # logger.info('%s Selection length %d should be %d.' %
                #        (name, len(rowIndex), len(n)))
                return {n: [x for x, s in zip(c.getData(), rowIndex) if s] for n, c in d.items()}
        else:
            raise ValueError(
                'RowIndex must be an int, a slice, or a list of ints or bools.')

    def getRow(self, rowIndex):
        """ Returns a list containing the objects located in a particular row, or a list of rows.

        rowIndex: ref ``getRowMap()``
* int: return the int-th row in a list of elements;
* ``Slice`` object, list of integers, list of booleans: return a list of rows each represented by a tuple. Example ``a.getRow(Slice(3,,))``, ``[2,4]``, ``[True, False...]``.
        """

        it = self.getRowMap(rowIndex).values()
        if issubclass(rowIndex.__class__, int):
            # return a list of row elements
            return list(it)
        # return transposed in a list
        return list(zip(*it))

    def select(self, selection):
        """ Select a number of rows from this table dataset and
        return a new TableDataset object containing only the selected rows.

        selection:  to form a new Tabledataset with ref ``getRowMap()``
        """

        d = ODict()
        if issubclass(selection.__class__, int):
            for name, data in self.getRowMap(selection).items():
                d[name] = Column(
                    data=[data], unit=self.getColumn(name).getUnit())
                return TableDataset(data=d)

        from .arraydataset import Column
        for name, data in self.getRowMap(selection).items():
            d[name] = Column(data=data, unit=self.getColumn(name).getUnit())
        return TableDataset(data=d)

    def removeRow(self, rowIndex):
        """ Removes a row with specified index from this table.

        rowIndex: int or a ``Slice`` object. Example ``a.removeRow(Slice(3,,))``.
        return: removed row data.
        """
        if issubclass(rowIndex.__class__, slice):
            ret = []
            for x in self.getData().values():
                ret.append(x.data[rowIndex])
                del x.data[rowIndex]
            return ret

        return [x.data.pop(rowIndex) for x in self.getData().values()]

    @ property
    def rowCount(self):
        return self.getRowCount()

    @ rowCount.setter
    def rowCount(self, newRowCount):
        self.setRowCount(newRowCount)

    def setRowCount(self, rowCount):
        """ cannot do this.
        """
        raise ValueError('Cannot set row count.')

    @ property
    def columnCount(self):
        return self.getColumnCount()

    @ columnCount.setter
    def columnCount(self, newColumnCount):
        self.setColumnCount(newColumnCount)

    def setColumnCount(self, columnCount):
        """ cannot do this.
        """
        raise ValueError('Cannot set column count.')

    @ property
    def list(self):
        return self._list

    @ list.setter
    def list(self, l):
        raise NotImplemented

    def getColumnMap(self, key=None):
        """ Returns a dict of column-names as the keys and the column(s) as the values.

        key: return the following as the value for each key-value pair:
* int: name-value where value is the int-th column.
* ``Slice`` object, a list of name-columns from slicing the column index. Example ``a.getColumn(Slice(3,,))``;
* Sequence of integers/strings: they are used as the column index/name to select the columns.
* Sequence of booleans: columns where the corresponding boolean is True are chosen.
Default is to return all columns.
        """
        d = self.getData()
        try:
            if d is None or key is None or len(d) == 0:
                return d
        except TypeError:
            pass

        cl = key.__class__
        if issubclass(cl, int):
            t = list(d.items())[key]
            return {t[0]: t[1]}       # {str:Column}
        if issubclass(cl, slice):
            return ODict(list(d.items())[key])  # {str:Column, ...}
        if issubclass(cl, str):
            return {key: d[key]}      # {str:Column}

        if issubclass(cl, Sequence):
            if type(key[0]) == int:
                # {str:Column, ...}
                return ODict(list(d.items())[i] for i in key)
            if type(key[0]) == str:
                return {n: d[n] for n in key}  # {str:Column, ...}
            if type(key[0]) == bool:
                # {str:Column, ...}
                return ODict(x for x, s in zip(d.items(), key) if s)
        else:
            raise ValueError(
                '``key`` must be an int, a string, a slice, or a list of ints, strings, or bools.')

    def getColumn(self, key):
        """ Returns the particular column, or a list of columns.

        key: ref ``getColumnMap()``
* int/str: return the int-th/named column;
* ``Slice`` object, list of columns of sliced column indices;
* list of integers/strings: return a list of columns corresponding to the given column index/name, or where key is True. Example ``a.getColumn(Slice(3,,))``, ``[2, 4]``, ``['time', ``energy']``.
* list of booleans: return a list of columns  where key is True. Example ``[True, False...]``.
        """

        it = self.getColumnMap(key).values()
        if len(it) == 0:
            return []
        if issubclass(key.__class__, (int, str)):
            # return a list of row elements
            return list(it)[0]
        # return transposed in a list
        return list(it)

    def setColumn(self, key, value):
        """ Replaces a column in this table with specified name to specified column if key is a string and exists, or if the key is an integer in 0 to the number of columns, insert at column-index=key, with the name 'column'+key, else add a new coolumn.
        """

        if self.getData() is None:
            key = ''
        elif type(key) == int:
            nms = self.getColumnNames()
            if 0 <= key and key < len(nms):
                key = nms[key]
            else:
                key = ''
        self.addColumn(name=key, column=value)

    def items(self):
        """ for k,v in tabledataset.items()
        """
        return self.getData().items()

    def __getitem__(self, key):
        """ return colmn if given key.

        ref. ``getColumn()``.
        """
        return self.getColumn(key)

    def __setitem__(self, key, value):
        """
        """
        self.setColumn(key, value)

    def __repr__(self):
        return self.toString(level=2)

    def toString(self, level=0,
                 tablefmt='rst', tablefmt1='simple', tablefmt2='simple',
                 width=0, param_widths=None, matprint=None, trans=True, **kwds):
        """
        tablefmt2: format of 2D data, others see `MetaData.toString`.
        """
        if matprint is None:
            matprint = ndprint

        cn = self.__class__.__name__
        s = '=== %s (%s) ===\n' % (cn, self.description if hasattr(
            self, 'descripion') else '')
        if level > 1:
            s = cn + '('
            s += self.meta.toString(
                level=level,
                tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                width=width, param_widths=param_widths,
                **kwds)
            return s + 'data= {' + \
                ', '.join('"%s": %s' % (k, v.toString(
                    level=level,
                    tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt,
                    width=width, param_widths=param_widths, **kwds))
                    for k, v in self.getColumnMap().items()) + \
                '})'
        else:
            s += mstr(self.__getstate__(), level=level,
                      tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                      excpt=['description'], **kwds)
        stp = 2 if level > 1 else 20 if level == 1 else None
        cols = self.getData().values()

        coldata = [list(itertools.islice(x.data, stp)) for x in cols]
        d = cn + '-dataset =\n'
        nmun = zip((str(x) for x in self.getData().keys()),
                   (str(x.unit) for x in cols))
        hdr = list('%s\n(%s)' % nu for nu in nmun)
        d += matprint(coldata, trans=trans, headers=hdr,
                      tablefmt=tablefmt, tablefmt1=tablefmt1,
                      tablefmt2=tablefmt2, **kwds)
        collen = self.getRowCount()
        if level and stp < collen:
            d += '(Only display %d rows of %d for level=%d.)' % (stp, collen, level)
        return s + '\n' + d + '\n'

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           meta=self.meta,
                           data=self.getData(),
                           _STID=self._STID)


class IndexedTableDataset(Indexed, TableDataset):
    """ TableDataset with an index table for efficient row look-up.

    """

    def __init__(self, **kwds):
        """
        """
        self._indexCols = [0]
        self._rowIndexTable = {}
        super().__init__(**kwds)  # initialize data, meta, unit

    def getColumnsToLookup(self):
        """ returns an iterator that gives a number of sequences to looking up over.
        """

        # list of Column's arrays
        return [x.data for x in self.getColumn(self._indexPattern)]

    def setData(self, data):
        """  sets name-column pairs from data and updates index if needed

        """

        d = self.getData()
        if d:
            reindex = False
            lcd = len(d)
            if issubclass(data.__class__, seqlist):
                for ind, x in enumerate(data):
                    if lcd > ind:
                        if reindex == False and ind in self._indexPattern:
                            reindex = True
        else:
            reindex = True
        super().setData(data)
        if reindex:
            self.updateToc()

    def vLookUp(self, key, return_index=True, multiple=False):
        """ Similar to Excel VLOOKUP, return all records (rows) that match the key.
        key: taken as a dictionary key unless ``multiple`` is True.
        return_index: if True (default) return index in the array of columns.
        multiple: if True (default is False) loop through key as a sequence of keys and return a sequece.
        """

        if multiple:
            if return_index:
                toc = self._tableOfContent
                return [toc[k] for k in key]
            else:
                toc = self._tableOfContent
                # return [[c[toc[k]] for c in self._list] for k in key]
                return list(zip(*((c[toc[k]] for k in key) for c in self._list)))
        else:
            if return_index:
                return self._tableOfContent[key]
            else:
                rec_ind = self._tableOfContent[key]
                return [c[rec_ind] for c in self._list]

    def hashx(self):
        s = self.__getstate__().values()
        l = []
        return super().hash(hash_list=self.data.values())

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return Indexed.__getstate__(self).update(description=self.description,
                                                 meta=self.meta,
                                                 data=self.getData(),
                                                 _STID=self._STID)



import nox


@nox.session(python=["3.6.8", "2.7"])
def testdataset(session):
    # session.install('pytest')
    session.install("-r", "requirements.txt")
    session.run('./test 1')


@nox.session(python=["3.6.8"])
def testpalpns(session):
    # session.install('pytest')
    session.install("-r", "requirements.txt")
    session.run('./test 2')
    session.run('./test 3')


#@nox.session
# def lint(session):
#    session.install('flake8')
#    session.run('flake8', '--import-order-style', 'google')



import paho.mqtt.client as mqtt
import json
import logging
import queue
import threading
import time

CONN_RESULT = {}
CONN_RESULT[0] = "Connection accepted"
CONN_RESULT[1] = "Connection Refused, unacceptable protocol version"
CONN_RESULT[2] = "Connection Refused, identifier rejected"
CONN_RESULT[3] = "Connection Refused, Server unavailable"
CONN_RESULT[4] = "Connection Refused, bad user name or password"
CONN_RESULT[5] = "Connection Refused, not authorized"


class queuework2:

    keepalive = 60
    logger = None
    subclient = None

    def __init__(self, topics,
                 host=None, port=None, username=None, passwd=None,
                 client_id=None,
                 callback=None,
                 qos=1,
                 userdata=None,
                 clean_session=None,
                 ):
        self.init_logger()
        self.init_args(topics,
                       host=host, port=port, username=username, passwd=passwd,
                       on_msg_callback=callback,
                       client_id=client_id, qos=qos, userdata=userdata,
                       clean_session=clean_session)
        self.client = None
        pass

    def init_logger(self):
        logging.basicConfig(
            level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger("queuework")
        pass

    def init_args(self, topics,
                  host=None, port=None, username=None, passwd=None,
                  client_id=None,
                  on_msg_callback=None,
                  qos=1, clean_session=None, userdata=None):
        self.topics = topics
        self._h = host
        self._p = port
        self.username = username
        self.passwd = passwd

        self.qos = qos
        self.client_id = client_id
        #
        self.msgq = queue.Queue()
        self.lock = threading.Lock()
        self.userdata = userdata
        self.on_msg_callback = on_msg_callback
        self.clean_session = clean_session
        self.reset = False
        # publish msg ids
        self.last_mid = -1
        self.mids_without_ack = []
        self.mwa_lock = threading.Lock()
        self.ordered_send = True

        pass

    def process_mid(self, mid, where):
        """ if mid is already in mwa, check it; if not put it in """

        mwa = self.mids_without_ack
        if mid and mid not in mwa:
            self.logger.debug(where + '%d not in MWA' % mid)
            self.last_mid = mid
            mwa.append(mid)
        else:
            if self.ordered_send:
                if mid > mwa[-1]:
                    self.logger.error(where + "Mid %d found before %d acknowledged." %
                                      (mid, mwa[-1]))
                elif mid < mwa[-1]:
                    self.logger.warning(where +
                                        "Mid %d acknowledged after %d has been sent." % (mid, mwa[-1]))
            mwa.remove(mid)
            self.logger.debug(where + 'MID %d removed from in MWA' % mid)
            msg = where + "Published with msg ID %d." % mid
            self.logger.debug(msg)

    def on_subscribe(self, client, userdata, mid, granted_qos):
        if client.subscr_mid == mid:
            client.subscribed = True

    def on_connect(self, client, userdata, flags, rc):
        msg = CONN_RESULT[rc]

        has_prev_data = ' some' if flags['session present'] else ' no'
        logmsg = "Connected with result: " + msg + \
            has_prev_data + ' prev session data.'
        if rc:
            client.connected = False
            self.logger.warning(logmsg)
        else:
            client.connected = True
            self.logger.debug(logmsg)
        pass

    def on_disconnect(self, client, userdata, rc):
        client.connected = False
        if rc != 0:
            msg = CONN_RESULT[rc] + " Unexpected disconnection."
        else:
            msg = CONN_RESULT[rc]
        self.logger.debug("Disconnected with result: " + msg)
        pass

    def on_message(self, client, userdata, msg):
        self.lock.acquire()
        self.msgq.put(msg)
        self.logger.debug(msg)
        self.lock.release()
        pass

    def on_publish(self, client, userdata, mid):
        self.mwa_lock.acquire()
        self.process_mid(mid, 'on_publish: ')
        self.mwa_lock.release()

    def wait_for(self, client, msgType, period=0.1):
        """ Waiting for subscribe and publish to clear.

        http://www.steves-internet-guide.com/subscribing-topics-mqtt-client/
        return: 0 if successful.
        """
        if msgType == "SUBACK":
            if client.on_subscribe:
                limit = 0
                while not client.subscribed:
                    self.logger.debug("waiting suback")
                    client.loop(period)  # check for messages
                    # time.sleep(period)
                    limit += period
                    if limit > 10:
                        self.logger.warning(msgType + " timeout.")
                        return 1
            return 0

        elif msgType == "CONNACK":
            if client.on_connect:
                limit = 0
                while not client.connected:
                    self.logger.debug("waiting connack")
                    client.loop(period)  # check for messages
                    # time.sleep(period)
                    limit += period
                    if limit > 10:
                        self.logger.warning(msgType + " timeout.")
                        return 1
                    # print(':::::::::::::')
            return 0

        elif msgType == "PUBACK":
            if client.on_publish:
                limit = 0
                while len(self.mids_without_ack):
                    self.logger.debug("waiting puback %s",
                                      str(self.mids_without_ack))
                    client.loop(period)  # check for messages
                    # time.sleep(period)
                    limit += period
                    if limit > 10:
                        self.logger.warning(msgType + " timeout.")
                        return 1
            return 0
        else:
            self.logger.error('Bad msgType ' + msgType)
            return 2

    def init_client(self, force=False):

        if self.client and not force:
            return self.client
        self.client = None
        # self.logger.debug(type(self.userdata))
        username = self.username
        password = self.passwd
        host = self._h
        port = self._p
        qos = self.qos

        # init
        client = mqtt.Client(client_id=self.client_id,
                             clean_session=self.clean_session,
                             userdata=self.userdata)
        self.client = client
        client.max_inflight_messages_set(100)
        client.on_connect = self.on_connect
        client.on_message = self.on_msg_callback if \
            self.on_msg_callback else self.on_message
        client.on_publish = self.on_publish
        client.on_disconnect = self.on_disconnect
        client.username_pw_set(username, password)

        # connect
        client.connect(host, port, self.keepalive)
        client.connected = False
        self.logger.debug("Connect " + host + ":" + str(port))
        # subscription ms ids
        client.subscr_mid = -1
        if self.wait_for(client, 'CONNACK'):
            return None

        # topic subscription
        if isinstance(self.topics, list):
            if isinstance(self.topics[0], str):
                # topics is a list of topics
                topics = [(topic, qos) for topic in self.topics]
            else:
                topics = self.topics
        elif isinstance(self.topics, str):
            topics = self.topics
        else:
            self.logger.error('Bad format for subscrib() topics ' +
                              str(self.topics))
            return None
        rc, mid = client.subscribe(topics, qos=qos)
        client.subscr_mid = mid
        self.logger.debug("subscribe %s status: %s mid %s" %
                          (str(topics), CONN_RESULT[rc], str(mid)))
        client.subscribed = False
        if self.wait_for(client, 'SUBACK'):
            return None

        return client

    def start_send(self):

        return self.init_client()

    def send(self, topics, text, conn=True):

        if conn:
            self.client.connect(self._h, self._p, self.keepalive)
            # self.client.reconnect()
            self.client.connected = False
            self.logger.debug("Connect " + self._h + ":" + str(self._p))
            if self.wait_for(self.client, 'CONNACK'):
                return 1
        #
        rc, mid = self.client.publish(topics, payload=text, qos=self.qos)
        self.mwa_lock.acquire()
        self.process_mid(mid, 'send: ')
        self.mwa_lock.release()
        self.logger.debug("Publish status: %d mid %d" % (rc, mid))
        if conn:
            if self.wait_for(self.client, 'PUBACK'):
                return 1
        # client.loop_stop()
        # client.disconnect()

        return 0

    def stop_send(self):
        if self.client is not None:
            # self.client.disconnect()
            # self.client.loop_start()

            # self.client = None
            pass

    def start_receive(self, loop='no'):

        if self.init_client():
            #
            if loop == 'forever':
                self.client.loop_forever()
            elif loop == 'start':
                self.client.loop_start()
            else:
                pass
        #

    def stop_receive(self):
        if self.subclient is not None:
            self.subclient.loop_stop()
            self.subclient.disconnect()
            self.subclient = None
        pass

    def get_size():
        size = self.msgq.qsize()
        return size

    def get_message(self):
        try:
            self.lock.acquire()
            # m = self.msgq.get(True,self.timeout)
            m = self.msgq.get_nowait()
        except Exception:
            m = None
        finally:
            self.lock.release()

        return m



# -*- coding: utf-8 -*-

from ..pns.config import pnsconfig as builtin_conf

from os.path import join, expanduser, expandvars, isdir
import functools
import sys
import importlib

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('logging level %d' % (logger.getEffectiveLevel()))


class Instance():

    def get(self, name=None, conf='pns'):
        if name:
            try:
                return self._cached_conf
            except AttributeError:
                self._cached_conf = getConfig(name=name, conf=conf)
                return self._cached_conf
        else:
            try:
                return self._cached_poolurl
            except AttributeError:
                self._cached_poolurl = getConfig(name=name, conf=conf)
                return self._cached_poolurl


# @functools.lru_cache(8)
def getConfig(name=None, conf='pns'):
    """ Imports a dict named [conf]config.

    The contents of the config are defined in the ``.config/[conf]local.py`` file. The contenss are used to update defaults in ``fdi.pns.config``.
    Th config file directory can be modified by the environment variable ``CONF_DIR``, which if  not given or pointing to an existing directly is process owner's ``~/.config`` directory.

    name: if given the poolurl in ``poolurl_of`` is returned, else construct a poolurl from the contents in dict <conf>config.
    conf: configuration ID. default 'pns', so the file is 'pnslocal.py'.
    """
    # default configuration is provided. Copy pns/config.py to ~/.config/pnslocal.py
    config = builtin_conf

    epath = expandvars('$CONF_DIR')
    if isdir(epath):
        confp = epath
    else:
        # environment variable CONFIG_DIR is not set
        env = expanduser(expandvars('$HOME'))
        # apache wsgi will return '$HOME' with no expansion
        if env == '$HOME':
            env = '/root'
        confp = join(env, '.config')
    # this is the var_name part of filename and the name of the returned dict
    var_name = conf+'config'
    module_name = conf+'local'
    file_name = module_name + '.py'
    filep = join(confp, file_name)
    absolute_name = importlib.util.resolve_name(module_name, None)
    logger.debug('Reading from configuration file %s/%s. absolute mod name %s' %
                 (confp, file_name, absolute_name))
    # if sys.path[0] != confp:
    #    sys.path.insert(0, confp)
    # print(sys.path)
    # for finder in sys.meta_path:
    #     spec = finder.find_spec(absolute_name, filep)
    #     print(spec)  # if spec is not None:

    try:
        spec = importlib.util.spec_from_file_location(absolute_name, filep)
        #print('zz', spec)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module_name] = module
        # the following suffers from non-updating loader
        # importlib.invalidate_caches()
        # module = importlib.import_module(module_name)
        # modul = __import__(module_name, globals(), locals(), [var_name], 0)
        config.update(getattr(module, var_name))
        logger.debug('Reading %s/%s done.' % (confp, file_name))
    except (ModuleNotFoundError, FileNotFoundError) as e:
        logger.warning(str(
            e) + '. Use default config in the package, such as fdi/pns/config.py. Copy it to ~/.config/[package]local.py and make persistent customization there.')

    if name:
        urlof = vars(c)['poolurl_of']
        if name in urlof:
            return urlof[name]
        else:
            return config['httphost'] + config['baseurl'] + '/' + name
    else:
        return config



# -*- coding: utf-8 -*-

from fdi.pal.poolmanager import PoolManager
from fdi.pns.jsonio import getJsonObj, postJsonObj, putJsonObj, commonheaders
from fdi.utils import getconfig
from fdi.utils.common import lls

import pytest
import os
import logging
logger = logging.getLogger(__name__)



# -*- coding: utf-8 -*-

from ..utils.getconfig import getConfig
from ..utils.common import lls
from .urn import parse_poolurl
from ..pns.fdi_requests import put_on_server, delete_from_server
from ..pal.productpool import ProductPool

from requests.exceptions import ConnectionError

import getpass
from weakref import WeakValueDictionary, getweakrefcount, finalize, getweakrefs
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

pc = getConfig()

DEFAULT_MEM_POOL = 'defaultmem'
# localpool
DEFAULT_POOL = 'fdi_pool_' + __name__ + getpass.getuser()


def remoteRegister(p, poolurl):
    logger.debug('Register %s on the server', poolurl)
    try:
        res, msg = put_on_server('urn:::0', poolurl, 'pool')
    except ConnectionError as e:
        res, msg = 'FAILED', str(e)
    if res == 'FAILED':
        raise RuntimeError(
            'Registering ' + poolurl + ' failed.  ' + msg)
    return res, msg


def remoteUnregister(poolurl):
    """ this method does not reference pool object. """
    if not poolurl.lower().startswith('http'):
        logger.warning('Ignored: %s not for a remote pool.' % poolurl)
        return 1
    logger.debug('unregister %s on the server', poolurl)
    #url = api_baseurl + post_poolid
    #x = requests.delete(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    #o = deserialize(x.text)
    urn = 'urn:::0'
    try:
        res, msg = delete_from_server(urn, poolurl, 'pool')
    except ConnectionError as e:
        res, msg = 'FAILED', str(e)
    if res == 'FAILED':
        logger.warning('Ignored: Unregisterinf ' +
                       poolurl + ' failed.  ' + msg)
        code = 2
    else:
        code = 0
    return code


class PoolManager(object):
    """
    This class provides the means to reference ProductPool objects without having to hard-code the type of pool. For example, it could be desired to easily switch from one pool type to another.

This is done by calling the getPool() method, which will return an existing pool or create a new one if necessary.
    """
    # Global centralized dict that returns singleton -- the same -- pool for the same ID.
    _GlobalPoolList = WeakValueDictionary()

    # maps scheme to default place/poolpath
    p = pc['node']['host']+':'+str(pc['node']['port'])+pc['baseurl']
    PlacePaths = {
        'file': pc['base_poolpath'],
        'mem': '/',
        'http': p,
        'https': p,
        'server': pc['server_poolpath'],
    }
    del p

    @classmethod
    def getPool(cls, poolname=None, poolurl=None, pool=None, **kwds):
        """ returns an instance of pool according to name or path of the pool.

        Returns the pool object if the pool is registered. Creates the pool if it does not already exist. the same poolname-path always get the same pool. Http pools will be registered on the sserver side.

Pools registered are kept as long as the last reference remains. When the last is gone the pool gets :meth;`removed` d.

        poolname: name of the pool.
        poolurl: if given the poolpath, scheme, place will be derived from it. if not given for making a new pool (i.e. when poolname is not a registered pool name.. 
If poolname is missing it is derived from poolurl; if poolurl is also absent, ValueError will be raised.
        kwds: passed to pool instanciation arg-list.

        """
        # logger.debug('GPL ' + str(id(cls._GlobalPoolList)) +
        #             str(cls._GlobalPoolList) + ' PConf ' + str(cls.PlacePaths))

        if pool:
            if poolname:
                raise ValueError(
                    'Pool name %s and pool object cannot be both given.' % poolname)
            poolname, poolurl, p = pool._poolname, pool._poolurl, pool
            if poolurl.lower().startswith('http'):
                res, msg = remoteRegister(p, poolurl)
        else:
            # find out the poolname first
            if poolname is None:
                if poolurl:
                    # the last segment will be the poolname
                    pp, schm, pl, poolname = parse_poolurl(poolurl)
            if cls.isLoaded(poolname):
                return cls._GlobalPoolList[poolname]

            if poolurl:
                # poolname use the one comes above
                pp, schm, pl, pn = parse_poolurl(poolurl)
            else:
                raise ValueError(
                    'A new pool %s cannot be created without a pool url.' % poolname)
            # now we have scheme, poolname, poolurl
            if schm == 'file':
                from . import localpool
                p = localpool.LocalPool(
                    poolname=poolname, poolurl=poolurl, **kwds)
            elif schm == 'mem':
                from . import mempool
                p = mempool.MemPool(poolname=poolname, poolurl=poolurl, **kwds)
            elif schm == 'server':
                from . import httppool
                p = httppool.HttpPool(
                    poolname=poolname, poolurl=poolurl, **kwds)
            elif schm in ('http', 'https'):
                from . import httpclientpool
                p = httpclientpool.HttpClientPool(
                    poolname=poolname, poolurl=poolurl, **kwds)
                res, msg = remoteRegister(p, poolurl)
            else:
                raise NotImplementedError(schm + ':// is not supported')
        #print(getweakrefs(p), id(p), '////')
        cls.save(poolname, p)
        #print(getweakrefs(p), id(p))

        # Pass poolurl to PoolManager.remove() for remote pools
        # finalize(p, print, poolname, poolurl)
        logger.debug('made pool ' + lls(p, 900))
        return p

    @ classmethod
    def getMap(cls):
        """
        Returns a poolname - poolobject map.
        """
        return cls._GlobalPoolList

    @ classmethod
    def isLoaded(cls, poolname):
        """
        Whether an item with the given id has been loaded (cached).

        returns the number of remaining week references if the pool is loaded.
        """
        if poolname in cls._GlobalPoolList:
            # print(poolname, getweakrefcount(cls._GlobalPoolList[poolname]))
            return getweakrefcount(cls._GlobalPoolList[poolname])
        else:
            return False

    @ classmethod
    def removeAll(cls):
        """ deletes all pools from the pool list, pools unwiped
        """
        nl = list(cls._GlobalPoolList)
        for pool in nl:
            cls.remove(pool)

    @ classmethod
    def save(cls, poolname, poolobj):
        """
        """
        cls._GlobalPoolList[poolname] = poolobj

    @ classmethod
    def remove(cls, poolname):
        """ Remove from list and unregister remote pools.

        returns 0 for successful removal, 1 for poolname not registered so no need, 2 for remove with warning.
        """
        l = cls.isLoaded(poolname)
        # print(getweakrefs(cls._GlobalPoolList[poolname]), id(
        #    cls._GlobalPoolList[poolname]), '......', l)
        if l is False:
            return 1
        code = 0
        if l == 1:
            # this is the only reference. unregister remote first.
            poolurl = cls._GlobalPoolList[poolname]._poolurl
            if poolurl.lower().startswith('http'):
                code = remoteUnregister(poolurl)
        del cls._GlobalPoolList[poolname]
        return code

    @ classmethod
    def getPoolurlMap(cls):
        """
        Gives the default poolurls of PoolManager.
        """
        return cls.PlacePaths

    @ classmethod
    def setPoolurlMap(cls, new):
        """
        Sets the default poolurls of PoolManager.
        """
        cls.PlacePaths.clear()
        cls.PlacePaths.update(new)

    @ classmethod
    def size(cls):
        """
        Gives the number of entries in this manager.
        """
        return len(cls._GlobalPoolList)

    def items(self):
        """
        Returns map's items
        """
        return self._GlobalPoolList.items()

    def __setitem__(self, *args, **kwargs):
        """ sets value at key.
        """
        self._GlobalPoolList.__setitem__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        """ returns value at key.
        """
        return self._GlobalPoolList.__getitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        """ removes value and its key.
        """
        self._GlobalPoolList.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        """ size of data
        """
        return self._GlobalPoolList.__len__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """ returns an iterator
        """
        return self._GlobalPoolList.__iter__(*args, **kwargs)

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self._GlobalPoolList) + ')'
